from __future__ import annotations

from typing import TYPE_CHECKING

from sw_mc_lib.Component import Component
from sw_mc_lib.Components import CompositeWriteBoolean, CompositeWriteNumber

if TYPE_CHECKING:
    from sw_mc_builder.wire import Wire


class ComponentWrapper:

    def __init__(
        self,
        inner_component: Component,
        inputs: dict[str, Wire],
        component_id: int = -1,
    ):
        self.inner_component: Component = inner_component
        self.inputs: dict[str, Wire] = inputs
        self.component_id: int = component_id
        self.optimize: bool = True
        self.force_property: bool = False

    def to_component(self) -> Component:
        self.inner_component.component_id = self.component_id
        for name, wire in self.inputs.items():
            if wire.component_id is not None:
                input_ = wire.to_input()
                assert input_ is not None
                if name.startswith("channel"):
                    assert isinstance(
                        self.inner_component,
                        (CompositeWriteNumber, CompositeWriteBoolean),
                    )
                    self.inner_component.channel_inputs[int(name.split("_")[1])] = (
                        input_
                    )
                else:
                    setattr(self.inner_component, name, input_)
        return self.inner_component

    def __hash__(self) -> int:
        return hash(id(self))



import sys
from functools import cache

import tumfl

from ._utils import INCLUDE_PATHS, normalize_path


@cache
def verify_script(script: str, minify: bool = True) -> str:
    """
    Verify the given script for correctness.
    This is a placeholder function and should be implemented with actual verification logic.
    """
    try:
        parsed = tumfl.parse(script)
    except tumfl.error.TumflError as e:
        if isinstance(e, tumfl.error.ParserError):
            print(e.full_error)
        else:
            print(e)
        sys.exit(1)
    if minify:
        tumfl.minify(parsed)
        return tumfl.format(parsed, style=tumfl.formatter.MinifiedStyle)
    return script


@cache
def resolve_and_verify_script(script_path: str) -> str:
    """
    Resolve dependencies in the given script and verify its correctness.
    """
    try:
        ast = tumfl.resolve_recursive(normalize_path(script_path), INCLUDE_PATHS)
    except tumfl.error.TumflError as e:
        if isinstance(e, tumfl.error.ParserError):
            print(e.full_error)
        else:
            print(e)
        sys.exit(1)
    return tumfl.format(ast)



from __future__ import annotations

import re
from typing import Optional

from sw_mc_lib.Components import (
    AND,
    NAND,
    NOR,
    NOT,
    OR,
    XOR,
    Abs,
    Add,
    ArithmeticFunction8In,
    BooleanFunction8In,
    Clamp,
    CompositeWriteBoolean,
    CompositeWriteNumber,
    ConstantNumber,
    ConstantOn,
    Divide,
    Equal,
    GreaterThan,
    LessThan,
    Modulo,
    Multiply,
    NumericalSwitchbox,
    Subtract,
    Threshold,
    UpDownCounter,
)
from sw_mc_lib.Types import SignalType

import sw_mc_builder.components as comp
from sw_mc_builder import wire

from .component_wrapper import ComponentWrapper
from .pseudo_components import PseudoComponent, Unconnected


def wire_to_id(w: wire.Wire) -> str:
    # deterministic one way function to get a unique id for a wire
    return str(id(w))


def process_and(inputs: dict[str, wire.Wire]) -> Optional[tuple[str, set[wire.Wire]]]:
    input_a = inner_optimize_numerical_boolean(inputs["a_input"])
    input_b = inner_optimize_numerical_boolean(inputs["b_input"])
    if input_a is None or input_b is None:
        return None
    true_a, used_wires_a = input_a
    true_b, used_wires_b = input_b
    true_func = f"({true_a}*{true_b})"
    return true_func, used_wires_a.union(used_wires_b)


def process_or(inputs: dict[str, wire.Wire]) -> Optional[tuple[str, set[wire.Wire]]]:
    input_a = inner_optimize_numerical_boolean(inputs["a_input"])
    input_b = inner_optimize_numerical_boolean(inputs["b_input"])
    if input_a is None or input_b is None:
        return None
    true_a, used_wires_a = input_a
    true_b, used_wires_b = input_b
    true_func = f"max({true_a},{true_b})"
    return true_func, used_wires_a.union(used_wires_b)


def process_xor(inputs: dict[str, wire.Wire]) -> Optional[tuple[str, set[wire.Wire]]]:
    input_a = inner_optimize_numerical_boolean(inputs["a_input"])
    input_b = inner_optimize_numerical_boolean(inputs["b_input"])
    if input_a is None or input_b is None:
        return None
    true_a, used_wires_a = input_a
    true_b, used_wires_b = input_b
    true_func = f"abs({true_a}-{true_b})"
    return true_func, used_wires_a.union(used_wires_b)


def process_not(inputs: dict[str, wire.Wire]) -> Optional[tuple[str, set[wire.Wire]]]:
    input_ = inner_optimize_numerical_boolean(inputs["a_input"])
    if input_ is None:
        return None
    true_func, used_wires = input_
    return f"(1-{true_func})", used_wires


def get_input(
    inputs: dict[str, wire.Wire], name: str, result_wires: set[wire.Wire]
) -> str:
    input_value = inputs[name]
    if isinstance(input_value.producer, ComponentWrapper):
        if isinstance(input_value.producer.inner_component, ConstantNumber):
            return input_value.producer.inner_component.value_property.text
    if isinstance(input_value.producer, Unconnected):
        return "0"
    result_wires.add(input_value)
    return wire_to_id(input_value)


def inner_optimize_numerical_boolean(
    input_wire: wire.Wire,
    first_iteration: bool = False,
) -> Optional[tuple[str, set[wire.Wire]]]:
    # pylint: disable=too-many-locals
    if isinstance(input_wire.producer, Unconnected):
        return "0", set()
    if not isinstance(input_wire.producer, ComponentWrapper):
        if first_iteration:
            return None
        switch_res = comp.numerical_switchbox(1, 0, input_wire)
        return wire_to_id(switch_res), {switch_res}
    component: ComponentWrapper = input_wire.producer
    if not component.optimize:
        return None
    inner = component.inner_component
    inputs = component.inputs
    result_wires: set[wire.Wire] = set()
    if isinstance(inner, Equal):
        epsilon = inner.epsilon_property.value
        a = get_input(inputs, "a_input", result_wires)
        b = get_input(inputs, "b_input", result_wires)
        true_func: str = f"((1-sgn(abs({a}-{b})-{epsilon}))/2)"
        if epsilon == 0:
            true_func = f"((sgn({a}-{b})+sgn({b}-{a}))/2)"
        return true_func, result_wires
    if isinstance(inner, GreaterThan):
        a = get_input(inputs, "a_input", result_wires)
        b = get_input(inputs, "b_input", result_wires)
        true_func = f"((1-sgn({b}-{a}))/2)"
        return true_func, result_wires
    if isinstance(inner, LessThan):
        a = get_input(inputs, "a_input", result_wires)
        b = get_input(inputs, "b_input", result_wires)
        true_func = f"((1-sgn({a}-{b}))/2)"
        return true_func, result_wires
    if isinstance(inner, Threshold):
        max_ = inner.max_property.value
        min_ = inner.min_property.value
        a = get_input(inputs, "a_input", result_wires)
        true_func = f"(((sgn({a}-{min_})+1)/2)*((sgn({max_}-{a})+1)/2))"
        return true_func, result_wires
    if isinstance(inner, NOT):
        input_ = inner_optimize_numerical_boolean(component.inputs["a_input"])
        if input_ is None:
            return None
        true_func, used_wires = input_
        return f"(1-{true_func})", used_wires
    if isinstance(inner, AND):
        return process_and(component.inputs)
    if isinstance(inner, NAND):
        and_result = process_and(component.inputs)
        if and_result is None:
            return None
        return f"(1-{and_result[0]})", and_result[1]
    if isinstance(inner, OR):
        return process_or(component.inputs)
    if isinstance(inner, NOR):
        or_result = process_or(component.inputs)
        if or_result is None:
            return None
        return f"(1-{or_result[0]})", or_result[1]
    if isinstance(inner, XOR):
        return process_xor(component.inputs)
    if isinstance(inner, ConstantOn):
        return "1", set()
    if first_iteration:
        return None
    switch_res = comp.numerical_switchbox(1, 0, input_wire)
    return wire_to_id(switch_res), {switch_res}


def optimize_numerical_boolean(
    input_wire: wire.Wire,
    extra_count: int = 0,
) -> Optional[tuple[str, dict[str, wire.Wire], list[str]]]:
    optimization_result = inner_optimize_numerical_boolean(
        input_wire, first_iteration=True
    )
    if optimization_result is None:
        return None
    true_func, used_wires = optimization_result
    if len(used_wires) + extra_count > 8:
        return None
    available_names = ["x", "y", "z", "w", "a", "b", "c", "d"]
    wire_associations: dict[str, wire.Wire] = {}
    for wire_ in used_wires:
        name = available_names.pop(0)
        true_func = true_func.replace(wire_to_id(wire_), name)
        wire_associations[name] = wire_
    return (
        true_func,
        wire_associations,
        [available_names.pop(0) for _ in range(extra_count)],
    )


def optimize_component(
    component: ComponentWrapper,
) -> Optional[ComponentWrapper | PseudoComponent]:
    inner = component.inner_component
    inputs = component.inputs
    if isinstance(inner, ArithmeticFunction8In):
        return comp.function(
            inner.function, **{k.replace("_input", ""): v for k, v in inputs.items()}
        ).producer
    if isinstance(inner, Add):
        return comp.function("x+y", inputs["a_input"], inputs["b_input"]).producer
    if isinstance(inner, Subtract):
        return comp.function("x-y", inputs["a_input"], inputs["b_input"]).producer
    if isinstance(inner, Multiply):
        return comp.function("x*y", inputs["a_input"], inputs["b_input"]).producer
    if isinstance(inner, Divide):
        return comp.function("x/y", inputs["a_input"], inputs["b_input"]).producer
    if isinstance(inner, Modulo):
        return comp.function("x%y", inputs["a_input"], inputs["b_input"]).producer
    if isinstance(inner, Abs):
        return comp.function("abs(x)", inputs["number_input"]).producer
    if isinstance(inner, Clamp):
        return comp.function(
            f"clamp(x,{inner.min_property.text},{inner.max_property.text})",
            inputs["number_input"],
        ).producer
    if isinstance(inner, ConstantNumber):
        return comp.function(inner.value_property.text).producer
    if isinstance(inner, BooleanFunction8In):
        return comp.boolean_function(
            inner.function, **{k.replace("_input", ""): v for k, v in inputs.items()}
        ).producer
    if isinstance(inner, ConstantOn):
        return comp.boolean_function("true").producer
    if isinstance(inner, AND):
        return comp.boolean_function(
            "x&y", inputs["a_input"], inputs["b_input"]
        ).producer
    if isinstance(inner, NAND):
        return comp.boolean_function(
            "!(x&y)", inputs["a_input"], inputs["b_input"]
        ).producer
    if isinstance(inner, OR):
        return comp.boolean_function(
            "x|y", inputs["a_input"], inputs["b_input"]
        ).producer
    if isinstance(inner, NOR):
        return comp.boolean_function(
            "!(x|y)", inputs["a_input"], inputs["b_input"]
        ).producer
    if isinstance(inner, XOR):
        return comp.boolean_function(
            "x^y", inputs["a_input"], inputs["b_input"]
        ).producer
    if isinstance(inner, NOT):
        return comp.boolean_function("!x", inputs["a_input"]).producer
    return None


def get_input_count(component: ComponentWrapper) -> int:
    return sum(
        0 if isinstance(input_wire.producer, Unconnected) else 1
        for input_wire in component.inputs.values()
    )


class Optimizer:
    def __init__(self, components: list[ComponentWrapper]):
        self.components: list[ComponentWrapper] = components
        self.optimized_components: dict[ComponentWrapper, ComponentWrapper] = {}
        self.to_visit: set[ComponentWrapper] = set(components)
        self.in_progress: set[ComponentWrapper] = set()

    def find_optimizations(self, component: ComponentWrapper) -> ComponentWrapper:
        # pylint: disable=too-many-locals,too-many-boolean-expressions
        if component in self.optimized_components:
            return self.optimized_components[component]

        self.in_progress.add(component)

        if not component.optimize:
            self.optimized_components[component] = component
            self.in_progress.remove(component)
            return component

        input_count = get_input_count(component)

        replacements: dict[str, str] = {}
        input_replacements: dict[str, wire.Wire] = {}
        present_inputs: dict[wire.Wire, str] = {}
        available_names: list[str] = ["x", "y", "z", "w", "a", "b", "c", "d"]
        potential_optimizations: list[tuple[int, str, ComponentWrapper, wire.Wire]] = []

        optimized = optimize_component(component)

        if optimized is None:
            inputs = component.inputs
            if isinstance(component.inner_component, NumericalSwitchbox):
                # try to optimize switch signal
                on_prod = inputs["on_value_input"]
                off_prod = inputs["off_value_input"]
                optimization_result = optimize_numerical_boolean(
                    inputs["switch_signal_input"], 2
                )
                if optimization_result is not None:
                    true_func, wire_associations, [on_name, off_name] = (
                        optimization_result
                    )
                    false_func = f"(1-{true_func})"
                    optimized = comp.function(
                        f"({true_func})*{on_name}+({false_func})*{off_name}",
                        **{on_name: on_prod, off_name: off_prod, **wire_associations},
                    ).producer
            elif isinstance(component.inner_component, UpDownCounter):
                up_optimized = inner_optimize_numerical_boolean(inputs["up_input"])
                down_optimized = inner_optimize_numerical_boolean(inputs["down_input"])
                reset_optimized = inner_optimize_numerical_boolean(
                    inputs["reset_input"]
                )
                min_value = component.inner_component.min_property.value
                max_value = component.inner_component.max_property.value
                increment_value = component.inner_component.increment_property.value
                clamp_value = component.inner_component.clamp
                if (
                    up_optimized is not None
                    and down_optimized is not None
                    and reset_optimized is not None
                ):
                    up_func, up_wires = up_optimized
                    down_func, down_wires = down_optimized
                    reset_func, reset_wires = reset_optimized
                    all_wires = up_wires.union(down_wires).union(reset_wires)
                    if len(all_wires) <= 7:  # need one extra for self reference
                        available_names = ["x", "y", "z", "w", "a", "b", "c", "d"]
                        wire_associations = {}
                        for wire_ in all_wires:
                            name = available_names.pop(0)
                            up_func = up_func.replace(wire_to_id(wire_), name)
                            down_func = down_func.replace(wire_to_id(wire_), name)
                            reset_func = reset_func.replace(wire_to_id(wire_), name)
                            wire_associations[name] = wire_
                        result = comp.placeholder(SignalType.Number)
                        self_name: str = available_names.pop(0)
                        wire_associations[self_name] = result
                        value = f"({self_name}+{increment_value}*{up_func}-{increment_value}*{down_func})"
                        if clamp_value:
                            value = f"clamp({value},{min_value},{max_value})"
                        result.replace_producer(
                            comp.function(
                                f"(1-{reset_func})*{value}+{reset_func}*{min_value}",
                                **wire_associations,
                            )
                        )
                        optimized = result.producer
            elif (
                isinstance(
                    component.inner_component,
                    (CompositeWriteBoolean, CompositeWriteNumber),
                )
                and component.inner_component.start_channel_property == 1
                and (other := component.inputs.get("composite_signal_input"))
                is not None
                and isinstance(other.producer, ComponentWrapper)
                and isinstance(
                    other.producer.inner_component, type(component.inner_component)
                )
                and isinstance(
                    other.producer.inner_component,
                    (CompositeWriteBoolean, CompositeWriteNumber),
                )
                and other.producer.inner_component.start_channel_property == 1
                and other.producer.optimize
            ):
                for i in range(1, 33):
                    name = f"channel_{i}_input"
                    if isinstance(component.inputs[name].producer, Unconnected):
                        component.inputs[name] = other.producer.inputs[name]
                component.inputs["composite_signal_input"] = other.producer.inputs[
                    "composite_signal_input"
                ]
                optimized = component

            assert optimized is None or isinstance(optimized, ComponentWrapper)
            self.optimized_components[component] = (
                component if optimized is None else self.find_optimizations(optimized)
            )
            if component in self.in_progress:
                self.in_progress.remove(component)
            return component if optimized is None else optimized

        assert isinstance(optimized, ComponentWrapper)
        assert isinstance(
            optimized.inner_component, (ArithmeticFunction8In, BooleanFunction8In)
        )
        self_references: set[str] = {
            name
            for name, input_wire in optimized.inputs.items()
            if input_wire.producer is component
        }

        for name, input_wire in optimized.inputs.items():
            raw_optimized = None
            if (
                isinstance(input_wire.producer, ComponentWrapper)
                and input_wire.producer not in self.in_progress
            ):
                raw_optimized = self.find_optimizations(input_wire.producer)
            if (
                isinstance(input_wire.producer, ComponentWrapper)
                and input_wire.producer.optimize
                and input_wire.producer not in self.in_progress
                and raw_optimized is not None
                and not any(
                    w.producer is raw_optimized for w in raw_optimized.inputs.values()
                )
            ):
                optimized_input = optimize_component(raw_optimized)
                if optimized_input is not None:
                    assert isinstance(optimized_input, ComponentWrapper)
                    potential_optimizations.append(
                        (
                            get_input_count(optimized_input) - 1,
                            name,
                            optimized_input,
                            input_wire,
                        )
                    )
                    continue
            if not isinstance(input_wire.producer, Unconnected):
                # keep input as is, and remove from available names
                shortened_name = name.replace("_input", "")
                available_names.remove(shortened_name)
                present_inputs[input_wire] = shortened_name
            else:
                # set unconnected inputs to 0, so that future assignments to these names work
                optimized.inner_component.function = re.sub(
                    rf"\b{name.replace('_input', '')}\b",
                    "0" if input_wire.wire_type == SignalType.Number else "false",
                    optimized.inner_component.function,
                )

        potential_optimizations.sort(key=lambda x: x[0])

        while sum(i[0] for i in potential_optimizations) + input_count > 8:
            # remove the optimization that saves the most inputs
            _, name, _, _ = potential_optimizations.pop()
            available_names.remove(name.replace("_input", ""))

        if not potential_optimizations:
            self.optimized_components[component] = component
            self.in_progress.remove(component)
            return component

        for _, name, optimized_input, input_wire in potential_optimizations:
            assert isinstance(
                optimized_input.inner_component,
                (ArithmeticFunction8In, BooleanFunction8In),
            )
            # rename variables in function to placeholder names (w_input_a, w_input_b, ...)
            function: str = optimized_input.inner_component.function
            for inner_name, inner_wire in optimized_input.inputs.items():
                var: str = inner_name.replace("_input", "")
                if not isinstance(inner_wire.producer, Unconnected):
                    if inner_wire in present_inputs:
                        input_count -= 1
                    new_name: str = f"{name}_{var}"
                    # remember new input
                    input_replacements[new_name] = inner_wire
                    function = re.sub(rf"\b{var}\b", new_name, function)
            # remember replacement in parent function
            replacements[name.replace("_input", "")] = f"({function})"
            # mark input as unconnected (can be reassigned later)
            optimized.inputs[name] = wire.Wire(input_wire.wire_type, Unconnected())

        new_function = optimized.inner_component.function
        # apply replacements
        for old, new in replacements.items():
            new_function = re.sub(rf"\b{old}\b", new, new_function)

        for name, new_wire in input_replacements.items():
            if new_wire in present_inputs:  # pylint: disable=consider-using-get
                # input already present, reuse name
                new_name = present_inputs[new_wire]
            else:
                new_name = available_names.pop(0)
            new_function = new_function.replace(name, new_name)
            optimized.inputs[f"{new_name}_input"] = new_wire
            if new_wire.producer is component:
                new_wire.producer = optimized
            present_inputs[new_wire] = new_name

        # if there are any self-references, update them to optimized
        for name in self_references:
            optimized.inputs[name].producer = optimized

        optimized.inner_component.function = new_function

        self.optimized_components[component] = optimized
        self.in_progress.remove(component)
        return optimized

    def apply(self) -> None:
        to_visit: set[ComponentWrapper] = set(self.components)
        visited: set[ComponentWrapper] = set()
        while to_visit:
            component = to_visit.pop()
            if component in visited:
                continue
            visited.add(component)
            for input_wire in component.inputs.values():
                if isinstance(input_wire.producer, ComponentWrapper):
                    input_wire.producer = self.find_optimizations(input_wire.producer)
                    to_visit.add(input_wire.producer)


def optimize_arithmetic(wires: list[wire.Wire]) -> None:
    """
    Optimize the given components by inlining arithmetic functions where possible.

    :param wires: List of ComponentWrapper to optimize
    :return: None
    """
    optimizer = Optimizer(
        [w.producer for w in wires if isinstance(w.producer, ComponentWrapper)]
    )
    for wire_ in wires:
        if isinstance(wire_.producer, ComponentWrapper):
            wire_.producer = optimizer.find_optimizations(wire_.producer)
    optimizer.apply()



from argparse import ArgumentParser


def parser_arguments(parser: ArgumentParser) -> None:
    parser.add_argument(
        "--microcontroller",
        "-m",
        action="store_true",
        help="Export microcontrollers to Stormworks microcontroller directory",
    )
    parser.add_argument(
        "--vehicle",
        "-v",
        type=str,
        help="Export microcontrollers to vehicles. Separated by commas.",
    )
    parser.add_argument(
        "--select",
        "-s",
        type=str,
        help="Select, which microcontrollers to export based on their name. Separated by commas.",
    )



import copy
import os
import sys
from argparse import ArgumentParser
from functools import cache
from pathlib import Path
from typing import Optional

from sw_mc_lib import XMLParserElement, format, parse
from sw_mc_lib.Types import ComponentType

from sw_mc_builder._handling_arguments import parser_arguments
from sw_mc_builder._utils import BUILDER_IDENTIFIER
from sw_mc_builder.microcontroller import Microcontroller


@cache
def get_stormworks_path() -> Path:
    # figure out platform-specific default Steam path
    stormworks_paths: list[Path] = []
    if sys.platform.startswith("win"):
        stormworks_paths = [Path(os.getenv("APPDATA")) / "Stormworks"]
    elif sys.platform.startswith("darwin"):  # macOS
        stormworks_paths = [Path.home() / "Library/Application Support/Stormworks"]
    else:  # Linux and others
        base_paths: set[Path] = {
            (Path.home() / ".steam/steam").resolve(),
            (Path.home() / ".local/share/Steam").resolve(),
        }
        for path in list(base_paths):
            libraryfolders_vdf = path / "steamapps/libraryfolders.vdf"
            if not libraryfolders_vdf.is_file():
                continue
            with libraryfolders_vdf.open() as f:
                while line := f.readline():
                    line = line.strip()
                    if line.startswith('"path"'):
                        # "path"\t\t"/some/path"
                        base_paths.add(Path(line.split("\t")[2].strip('"')).resolve())

        for path in base_paths:
            stormworks_paths.append(
                path
                / "steamapps/compatdata/573090/pfx/drive_c/users/steamuser/AppData/Roaming/Stormworks/"
            )

    stormworks_paths = [p for p in stormworks_paths if p.is_dir()]

    if len(stormworks_paths) == 0:
        raise FileNotFoundError("Could not find Stormworks installation path.")
    if len(stormworks_paths) > 1:
        raise RuntimeError(
            f"Found multiple Stormworks installation paths: {stormworks_paths}"
        )
    return stormworks_paths[0]


def name_to_path(name: str, kind: str) -> Path:
    if name.endswith(".xml"):
        return Path(name).resolve()
    return get_stormworks_path() / f"data/{kind}/{name}.xml"


def extract_mcs_from_vehicle(
    vehicle_path: Path, relevant_names: set[str]
) -> Optional[tuple[XMLParserElement, list[XMLParserElement]]]:
    with vehicle_path.open("r", encoding="utf-8") as f:
        vehicle_xml: XMLParserElement = parse(f.read())
    if vehicle_xml.tag != "vehicle":
        return None
    bodies = vehicle_xml.children[1]
    if bodies.tag != "bodies":
        return None
    relevant_mcs: list[XMLParserElement] = []
    for body in bodies.children:
        if body.tag != "body":
            continue
        components = body.children[0]
        if components.tag != "components":
            continue
        for component in components.children:
            if component.tag != "c":
                continue
            if component.attributes.get("d") != "microprocessor":
                continue
            obj = component.get_child_by_tag("o")
            if obj is None:
                continue
            mc_definition = obj.get_child_by_tag("microprocessor_definition")
            if mc_definition is None:
                continue
            if mc_definition.attributes.get("name") not in relevant_names:
                continue
            group = mc_definition.get_child_by_tag("group")
            if group is None:
                continue
            data = group.get_child_by_tag("data")
            if data is not None and data.attributes.get("desc") == BUILDER_IDENTIFIER:
                relevant_mcs.append(mc_definition)
    return vehicle_xml, relevant_mcs


def find_with_id(
    element: XMLParserElement, component_type: str
) -> list[XMLParserElement]:
    assert element.tag in ("microprocessor", "microprocessor_definition")
    group = element.get_child_by_tag_strict("group")
    components = group.get_child_by_tag_strict("components")
    result: list[XMLParserElement] = []
    for component in components.children:
        if component.attributes.get("type") == component_type:
            result.append(component)
    return result


def find_and_match_property(
    target: XMLParserElement,
    source: XMLParserElement,
    component_type: ComponentType,
) -> list[tuple[XMLParserElement, XMLParserElement]]:
    target_props = find_with_id(target, str(component_type.value))
    source_props = find_with_id(source, str(component_type.value))
    results = []
    for target_prop in target_props:
        target_obj = target_prop.get_child_by_tag_strict("object")
        if target_obj.attributes.get("force_property"):
            continue  # skip properties that are forced to be changed
        prop_name = target_obj.attributes.get("name", target_obj.attributes.get("n"))
        assert prop_name is not None
        for source_prop in source_props:
            source_obj = source_prop.get_child_by_tag_strict("object")
            source_name = source_obj.attributes.get(
                "name", source_obj.attributes.get("n")
            )
            assert source_name is not None
            if prop_name == source_name:
                results.append((target_obj, source_obj))
                break
    return results


def merge_attributes(
    target: XMLParserElement, source: XMLParserElement, attr_name: str
) -> None:
    if attr_name in source.attributes:
        target.attributes[attr_name] = source.attributes[attr_name]
    elif attr_name in target.attributes:
        del target.attributes[attr_name]


def merge_number_properties(
    target: XMLParserElement, source: XMLParserElement, prop_name: str
) -> None:
    target_val = target.get_child_by_tag_strict(prop_name)
    if source_val := source.get_child_by_tag(prop_name):
        target_val.attributes = source_val.attributes


def merge_properties(target: XMLParserElement, source: XMLParserElement) -> None:
    for target_prop, source_prop in find_and_match_property(
        target, source, ComponentType.PropertyText
    ):
        merge_attributes(target_prop, source_prop, "v")
    for target_prop, source_prop in find_and_match_property(
        target, source, ComponentType.PropertyNumber
    ):
        merge_number_properties(target_prop, source_prop, "v")
    for target_prop, source_prop in find_and_match_property(
        target, source, ComponentType.PropertyToggle
    ):
        merge_attributes(target_prop, source_prop, "v")
    for target_prop, source_prop in find_and_match_property(
        target, source, ComponentType.PropertySlider
    ):
        merge_number_properties(target_prop, source_prop, "v")
    for target_prop, source_prop in find_and_match_property(
        target, source, ComponentType.PropertyDropdown
    ):
        possible_values: dict[str, str] = {}
        items = target_prop.get_child_by_tag_strict("items")
        for child in items.children:
            if child.tag == "i":
                possible_values[(child.attributes["l"])] = str(len(possible_values))
        if (source_items := source_prop.get_child_by_tag("items")) and (
            source_index := source_prop.attributes.get("i")
        ):
            if int(source_index) < len(source_items.children):
                source_item = source_items.children[int(source_index)]
                if source_item.tag == "i":
                    label = source_item.attributes.get("l")
                    if label in possible_values:
                        target_prop.attributes["i"] = possible_values[label]


def replace_in_vehicle(
    vehicle_name: str,
    mcs: dict[str, tuple[Microcontroller, XMLParserElement]],
) -> None:
    vehicle_path = name_to_path(vehicle_name, "vehicles")
    if not vehicle_path.is_file():
        print(f'Could not find vehicle file "{vehicle_name}".')
        return
    result = extract_mcs_from_vehicle(vehicle_path, set(mcs.keys()))
    if result is None:
        print(f'Could not parse vehicle file "{vehicle_name}".')
        return
    vehicle_xml, vehicle_mcs = result
    if len(vehicle_mcs) == 0:
        print(f'No microcontrollers to update in vehicle "{vehicle_name}".')
        return
    for vehicle_mc in vehicle_mcs:
        name: str = vehicle_mc.attributes["name"]
        mc_xml: XMLParserElement = copy.deepcopy(mcs[name][1])
        merge_properties(mc_xml, vehicle_mc)
        vehicle_mc.attributes = mc_xml.attributes
        vehicle_mc.children = mc_xml.children
    with vehicle_path.open("w", encoding="utf-8") as f:
        f.write(format(vehicle_xml))
    print(f'Wrote microcontrollers to vehicle "{vehicle_name}".')


def handle_mcs(*mcs: Microcontroller) -> None:
    # pylint: disable=protected-access
    if len(mcs) == 0:
        raise ValueError("At least one microcontroller must be provided")
    parser = ArgumentParser(description="Microcontroller collection")
    parser_arguments(parser)
    args = parser.parse_args()
    if args.select:
        selected_names = {name.strip() for name in args.select.split(",")}
    else:
        selected_names = set()
    compiled: dict[str, tuple[Microcontroller, XMLParserElement]] = {}
    for mc in mcs:
        name: str = mc._mc.name
        if selected_names and not any(
            selected_name in name for selected_name in selected_names
        ):
            continue
        if name in compiled:
            raise ValueError(f"Duplicate microcontroller name: {name}")
        compiled[name] = (mc, mc._resolve_and_optimize())
    if args.microcontroller:
        for name, (mc, xml_mc) in compiled.items():
            mc_path = name_to_path(mc._save_name, "microprocessors")
            mc_path.parent.mkdir(parents=True, exist_ok=True)
            with mc_path.open("w", encoding="utf-8") as f:
                f.write(format(xml_mc))
                print(f'Wrote microcontroller "{name}" to microcontroller directory.')
            mc._mc.image.to_sw_png(mc_path.with_suffix(".png"))
    if args.vehicle:
        for vehicle_name in args.vehicle.split(","):
            replace_in_vehicle(vehicle_name, compiled)



import runpy
import sys
from argparse import ArgumentParser
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from sw_mc_builder import _utils
from sw_mc_builder._handling_arguments import parser_arguments

BLANK_MC: str = """\
from sw_mc_builder import *

input1 = comp.input(SignalType.Number, "Input 1")
input2 = comp.input(SignalType.Number, "Input 2")

added = input1 + input2

mc = Microcontroller("Example MC")
mc.place_input(input1, 0, 0)
mc.place_input(input2, 0, 1)
mc.place_output(added, "Added", x=1, y=0)

if __name__ == "__main__":
    handle_mcs(mc)
"""


@contextmanager
def temporary_argv(new_argv: list[str]) -> Generator[None, None, None]:
    old = sys.argv
    sys.argv = list(new_argv)
    try:
        yield
    finally:
        sys.argv = old


def run_child_script(child_path: Path, args: list[str]) -> None:
    resolved_child_path = child_path.resolve()
    # argv[0] will be the script filename (often useful to child)
    argv: list[str] = [str(resolved_child_path), *args]
    with temporary_argv(argv):
        # As imports are not reevaluated, we need to reset the main path
        _utils.MAIN_PATH = resolved_child_path.parent
        _utils.INCLUDE_PATHS.clear()
        _utils.INCLUDE_PATHS.append(_utils.MAIN_PATH)
        if str(resolved_child_path.parent) not in sys.path:
            sys.path.insert(0, str(resolved_child_path.parent))
        # run as if executed directly: __name__ == "__main__"
        runpy.run_path(str(resolved_child_path), run_name="__main__")


def initialize_mc(name: Path) -> None:
    name = name.with_suffix(".py")
    if name.exists():
        print(f"Error: File {name} already exists.")
        sys.exit(1)
    with name.open("w") as file:
        file.write(BLANK_MC)
    print(f"Initialized new microcontroller project at {name}")


def main() -> None:
    parser = ArgumentParser()
    subparsers = parser.add_subparsers()
    init = subparsers.add_parser(
        "init", help="Initialize a new microcontroller project", aliases=["i"]
    )
    init.add_argument("name", type=Path, help="Name of the microcontroller project")
    init.set_defaults(func="init")
    run = subparsers.add_parser(
        "run", help="Execute a microcontroller project", aliases=["r"]
    )
    run.add_argument(
        "name", type=Path, help="Path to the microcontroller definition file"
    )
    parser_arguments(run)
    run.set_defaults(func="run")
    args = parser.parse_args()
    if not hasattr(args, "func"):
        # No subcommand provided
        parser.print_help()
        sys.exit(1)
    elif args.func == "init":
        initialize_mc(args.name)
    elif args.func == "run":
        child_args: list[str] = []
        if args.microcontroller:
            child_args.append("--microcontroller")
        if args.select:
            child_args.extend(["--select", args.select])
        if args.vehicle:
            child_args.append("--vehicle")
            child_args.extend(args.vehicle)
        run_child_script(args.name, child_args)


if __name__ == "__main__":
    main()



# pylint: disable=too-many-lines, too-many-arguments, too-many-locals, redefined-outer-name
from typing import Optional

from sw_mc_lib import Position
from sw_mc_lib.Components import *
from sw_mc_lib.Components.property_dropdown import DropDownOption
from sw_mc_lib.Types import PulseMode, SignalType, TimerUnit

from sw_mc_builder import script_handling

from .component_wrapper import ComponentWrapper
from .pseudo_components import InputPlaceholder, Placeholder, Unconnected
from .wire import (
    AudioInput,
    AudioWire,
    BooleanInput,
    BooleanWire,
    CompositeInput,
    CompositeWire,
    NumberInput,
    NumberWire,
    VideoInput,
    VideoWire,
    Wire,
)

Number = float | int


def input(
    signal_type: SignalType,
    name: str,
    description: str = "The input signal to be processed.",
) -> Wire:
    """An input node for the microcontroller. This represents an input from the vehicle."""
    return Wire(signal_type, InputPlaceholder(-1, name, description), 0)


def add(a: NumberInput, b: NumberInput) -> NumberWire:
    """Adds the two input values together and outputs the result."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Add(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def sub(a: NumberInput, b: NumberInput) -> NumberWire:
    """Subtracts the second input from the first and outputs the result."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Subtract(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def mul(a: NumberInput, b: NumberInput) -> NumberWire:
    """Multiplies the two input values and outputs the result."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Multiply(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def div(a: NumberInput, b: NumberInput) -> tuple[NumberWire, BooleanWire]:
    """Divides the first input by the second and outputs the result. Also outputs, if the division was divieded by zero."""
    comp = ComponentWrapper(
        Divide(0, Position()),
        {
            "a_input": Wire.to(SignalType.Number, a),
            "b_input": Wire.to(SignalType.Number, b),
        },
    )
    return Wire(SignalType.Number, comp), Wire(SignalType.Boolean, comp, node_index=1)


def mod(a: NumberInput, b: NumberInput) -> NumberWire:
    """Outputs the modulo of input A by input B."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Modulo(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def abs(a: NumberInput) -> NumberWire:
    """Outputs the absolute value of the input value (negative numbers become positive)."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Abs(0, Position()),
            {"number_input": Wire.to(SignalType.Number, a)},
        ),
    )


def clamp(a: NumberInput, min_value: Number, max_value: Number) -> NumberWire:
    """Clamps the input value between a set min and max and outputs the result."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Clamp(0, Position(), min_value, max_value),
            {"number_input": Wire.to(SignalType.Number, a)},
        ),
    )


def constant_number(value: Number) -> NumberWire:
    """Outputs a constant number that is set as property."""
    return Wire(
        SignalType.Number, ComponentWrapper(ConstantNumber(0, Position(), value), {})
    )


def delta(value: NumberInput) -> NumberWire:
    """Delta Of Input Value. Outputs the difference between the current and previous input value."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            Delta(0, Position()),
            {"value_input": Wire.to(SignalType.Number, value)},
        ),
    )


def equal(
    a: NumberInput = None, b: NumberInput = None, epsilon: Optional[Number] = None
) -> BooleanWire:
    """Compares whether or not two numbers are equal within a set accuracy."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            Equal(0, Position(), epsilon_property=epsilon),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def function(
    function_string: str,
    x: NumberInput = None,
    y: NumberInput = None,
    z: NumberInput = None,
    w: NumberInput = None,
    a: NumberInput = None,
    b: NumberInput = None,
    c: NumberInput = None,
    d: NumberInput = None,
) -> NumberWire:
    """Evaluates a mathematical expression with up to 8 input variables and outputs the result."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            ArithmeticFunction8In(0, Position(), function_string),
            {
                "x_input": Wire.to(SignalType.Number, x),
                "y_input": Wire.to(SignalType.Number, y),
                "z_input": Wire.to(SignalType.Number, z),
                "w_input": Wire.to(SignalType.Number, w),
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
                "c_input": Wire.to(SignalType.Number, c),
                "d_input": Wire.to(SignalType.Number, d),
            },
        ),
    )


def unconnected(signal_type: SignalType) -> Wire:
    """
    Unconnected component for unused inputs. Does not have to be resolved to generate a valid microcontroller.
    """
    return Wire(signal_type, Unconnected())


def placeholder(signal_type: SignalType) -> Wire:
    """
    Placeholder component for incomplete circuits.
    Can be used for circular dependencies.
    Must be resolved to generate a valid microcontroller.
    """
    return Wire(signal_type, Placeholder())


def and_(a: BooleanInput, b: BooleanInput) -> BooleanWire:
    """Outputs the logical AND of its two input signals."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            AND(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
            },
        ),
    )


def or_(a: BooleanInput, b: BooleanInput) -> BooleanWire:
    """Outputs the logical OR of its two input signals."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            OR(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
            },
        ),
    )


def xor(a: BooleanInput, b: BooleanInput) -> BooleanWire:
    """Outputs the logical XOR of its two input signals."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            XOR(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
            },
        ),
    )


def nand(a: BooleanInput, b: BooleanInput) -> BooleanWire:
    """Outputs the logical NAND of its two input signals."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            NAND(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
            },
        ),
    )


def nor(a: BooleanInput, b: BooleanInput) -> BooleanWire:
    """Outputs the logical NOR of its two input signals."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            NOR(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
            },
        ),
    )


def not_(a: BooleanInput) -> BooleanWire:
    """Outputs the logical NOT of its input signal."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            NOT(0, Position()),
            {
                "a_input": Wire.to(SignalType.Boolean, a),
            },
        ),
    )


def constant_on() -> BooleanWire:
    """Outputs a constant on signal."""
    return Wire(SignalType.Boolean, ComponentWrapper(ConstantOn(0, Position()), {}))


def jk_flip_flop(
    set_: BooleanInput = None, reset: BooleanInput = None
) -> tuple[BooleanWire, BooleanWire]:
    """
    An JK flip flop that can be set and reset using two on/off inputs.
    Returns (Q, not Q).
    """
    component = ComponentWrapper(
        JKFlipFlop(0, Position()),
        {
            "set_input": Wire.to(SignalType.Boolean, set_),
            "reset_input": Wire.to(SignalType.Boolean, reset),
        },
    )
    return Wire(SignalType.Boolean, component), Wire(
        SignalType.Boolean, component, node_index=1
    )


def sr_latch(
    set_: BooleanInput = None, reset: BooleanInput = None
) -> tuple[BooleanWire, BooleanWire]:
    """
    An SR latch that can be set and reset using two on/off inputs.
    Returns (Q, not Q).
    """
    component = ComponentWrapper(
        SRLatch(0, Position()),
        {
            "set_input": Wire.to(SignalType.Boolean, set_),
            "reset_input": Wire.to(SignalType.Boolean, reset),
        },
    )
    return Wire(SignalType.Boolean, component), Wire(
        SignalType.Boolean, component, node_index=1
    )


def boolean_function(
    function_string: str,
    x: BooleanInput = None,
    y: BooleanInput = None,
    z: BooleanInput = None,
    w: BooleanInput = None,
    a: BooleanInput = None,
    b: BooleanInput = None,
    c: BooleanInput = None,
    d: BooleanInput = None,
) -> BooleanWire:
    """Evaluates a logical expression with up to 8 input variables and outputs the result."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            BooleanFunction8In(0, Position(), function_string),
            {
                "x_input": Wire.to(SignalType.Boolean, x),
                "y_input": Wire.to(SignalType.Boolean, y),
                "z_input": Wire.to(SignalType.Boolean, z),
                "w_input": Wire.to(SignalType.Boolean, w),
                "a_input": Wire.to(SignalType.Boolean, a),
                "b_input": Wire.to(SignalType.Boolean, b),
                "c_input": Wire.to(SignalType.Boolean, c),
                "d_input": Wire.to(SignalType.Boolean, d),
            },
        ),
    )


def pulse(
    toggle: BooleanInput = None, mode: PulseMode = PulseMode.OffToOn
) -> BooleanWire:
    """
    A switch that outputs a single tick pulse.
    It can be configured to pulse when being switched from off to on (default), on to off,
    or always when the input signal changes.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            Pulse(0, Position(), mode_property=mode),
            {"toggle_signal_input": Wire.to(SignalType.Boolean, toggle)},
        ),
    )


def push_to_toggle(toggle_signal: BooleanInput = None) -> BooleanWire:
    """An on/off switch that is toggled every time a new on signal is sent to its input."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            PushToToggle(0, Position()),
            {"a_input": Wire.to(SignalType.Boolean, toggle_signal)},
        ),
    )


def blinker(
    control_signal: BooleanInput = None,
    blink_on_duration: Number = 1.0,
    blink_off_duration: Number = 1.0,
) -> BooleanWire:
    """Outputs a value that blinks between on and off at a set rate."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            Blinker(0, Position(), None, blink_on_duration, blink_off_duration),
            {"control_signal_input": Wire.to(SignalType.Boolean, control_signal)},
        ),
    )


def capacitor(
    charge: BooleanInput = None, charge_time: Number = 1.0, discharge_time: Number = 1.0
) -> BooleanWire:
    """Charges up when receiving an on signal, then discharges over a period of time."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            Capacitor(0, Position(), None, charge_time, discharge_time),
            {"charge_input": Wire.to(SignalType.Boolean, charge)},
        ),
    )


def greater_than(a: NumberInput = None, b: NumberInput = None) -> BooleanWire:
    """Outputs an on signal if the first input is greater than the second."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            GreaterThan(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def less_than(a: NumberInput = None, b: NumberInput = None) -> BooleanWire:
    """Outputs an on signal if the first input is less than the second."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            LessThan(0, Position()),
            {
                "a_input": Wire.to(SignalType.Number, a),
                "b_input": Wire.to(SignalType.Number, b),
            },
        ),
    )


def memory_register(
    set_: BooleanInput = None,
    reset: BooleanInput = None,
    data: NumberInput = None,
    reset_value: Number = 0,
) -> NumberWire:
    """
    Remembers the input value when receiving a signal to the Set node.
    When the Reset node receives a signal, the stored number is cleared to a value that can be customised in the property.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            MemoryRegister(0, Position(), reset_value),
            {
                "set_input": Wire.to(SignalType.Boolean, set_),
                "reset_input": Wire.to(SignalType.Boolean, reset),
                "number_to_store_input": Wire.to(SignalType.Number, data),
            },
        ),
    )


def numerical_junction(
    value: NumberInput = None, switch: BooleanInput = None
) -> tuple[NumberWire, NumberWire]:
    """
    Outputs the input number to one of the outputs depending on whether or not the Switch Signal is on.
    The path that the input doesn't take will output a value of 0.
    """
    component = ComponentWrapper(
        NumericalJunction(0, Position()),
        {
            "value_to_pass_through_input": Wire.to(SignalType.Number, value),
            "switch_signal_input": Wire.to(SignalType.Boolean, switch),
        },
    )
    return Wire(SignalType.Number, component), Wire(
        SignalType.Number, component, node_index=1
    )


def numerical_switchbox(
    on_value: NumberInput = None,
    off_value: NumberInput = None,
    switch: BooleanInput = None,
) -> NumberWire:
    """
    Outputs the first input value when receiving an on signal, and the second when receiving an off signal.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            NumericalSwitchbox(0, Position()),
            {
                "on_value_input": Wire.to(SignalType.Number, on_value),
                "off_value_input": Wire.to(SignalType.Number, off_value),
                "switch_signal_input": Wire.to(SignalType.Boolean, switch),
            },
        ),
    )


def pid(
    setpoint: NumberInput = None,
    process_variable: NumberInput = None,
    active: BooleanInput = None,
    kp: Number = 0.0,
    ki: Number = 0.0,
    kd: Number = 0.0,
) -> NumberWire:
    """
    A basic PID controller. The proportional, integral and derivative gains can be set using properties.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            PIDController(
                0,
                Position(),
                proportional_property=kp,
                integral_property=ki,
                derivative_property=kd,
            ),
            {
                "setpoint_input": Wire.to(SignalType.Number, setpoint),
                "process_variable_input": Wire.to(SignalType.Number, process_variable),
                "active_input": Wire.to(SignalType.Boolean, active),
            },
        ),
    )


def advanced_pid(
    setpoint: NumberInput = None,
    process_variable: NumberInput = None,
    proportional_gain: NumberInput = None,
    integral_gain: NumberInput = None,
    derivative_gain: NumberInput = None,
    active: BooleanInput = None,
) -> NumberWire:
    """
    A PID controller with variable proportional, integral and derivative gains.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            PIDControllerAdvanced(0, Position()),
            {
                "setpoint_input": Wire.to(SignalType.Number, setpoint),
                "process_variable_input": Wire.to(SignalType.Number, process_variable),
                "proportional_input": Wire.to(SignalType.Number, proportional_gain),
                "integral_input": Wire.to(SignalType.Number, integral_gain),
                "derivative_input": Wire.to(SignalType.Number, derivative_gain),
                "active_input": Wire.to(SignalType.Boolean, active),
            },
        ),
    )


def threshold(
    value: NumberInput = None,
    low_threshold: Number = 0.0,
    high_threshold: Number = 1.0,
) -> BooleanWire:
    """Outputs an on/off signal indicating whether or not the input value is within a set threshold."""
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            Threshold(0, Position(), low_threshold, high_threshold),
            {
                "number_input": Wire.to(SignalType.Number, value),
            },
        ),
    )


def timer_rtf(
    enable: BooleanInput = None,
    duration: NumberInput = None,
    reset: BooleanInput = None,
    duration_unit: TimerUnit = TimerUnit.Seconds,
) -> BooleanWire:
    """
    Variable input timer. Outputs an on signal when the timer is less than its duration. The timer will not reset until it is signalled.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            TimerRTF(0, Position(), duration_unit),
            {
                "timer_enable_input": Wire.to(SignalType.Boolean, enable),
                "duration_input": Wire.to(SignalType.Number, duration),
                "reset_input": Wire.to(SignalType.Boolean, reset),
            },
        ),
    )


def timer_rto(
    enable: BooleanInput = None,
    duration: NumberInput = None,
    reset: BooleanInput = None,
    duration_unit: TimerUnit = TimerUnit.Seconds,
) -> BooleanWire:
    """
    Variable input timer. Outputs an on signal when the timer reaches its duration. The timer will not reset until it is signalled.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            TimerRTO(0, Position(), duration_unit),
            {
                "timer_enable_input": Wire.to(SignalType.Boolean, enable),
                "duration_input": Wire.to(SignalType.Number, duration),
                "reset_input": Wire.to(SignalType.Boolean, reset),
            },
        ),
    )


def timer_tof(
    enable: BooleanInput = None,
    duration: NumberInput = None,
    duration_unit: TimerUnit = TimerUnit.Seconds,
) -> BooleanWire:
    """
    Variable input timer. Outputs an on signal when the timer is less than its duration. The timer will reset when off.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            TimerTOF(0, Position(), duration_unit),
            {
                "timer_enable_input": Wire.to(SignalType.Boolean, enable),
                "duration_input": Wire.to(SignalType.Number, duration),
            },
        ),
    )


def timer_ton(
    enable: BooleanInput = None,
    duration: NumberInput = None,
    duration_unit: TimerUnit = TimerUnit.Seconds,
) -> BooleanWire:
    """
    Variable input timer. Outputs an on signal when the timer reaches its duration. The timer will reset when off.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            TimerTON(0, Position(), duration_unit),
            {
                "timer_enable_input": Wire.to(SignalType.Boolean, enable),
                "duration_input": Wire.to(SignalType.Number, duration),
            },
        ),
    )


def up_down_counter(
    up: BooleanInput = None,
    down: BooleanInput = None,
    reset: BooleanInput = None,
    increment: Number = 1,
    reset_value: Number = 0,
    min_value: Number = 0,
    max_value: Number = 0,
    clamp: bool = False,
) -> NumberWire:
    """Has an internal value that will increase and decrease when receiving different signals."""
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            UpDownCounter(
                0,
                Position(),
                min_property=min_value,
                max_property=max_value,
                clamp=clamp,
                increment_property=increment,
                reset_property=reset_value,
            ),
            {
                "up_input": Wire.to(SignalType.Boolean, up),
                "down_input": Wire.to(SignalType.Boolean, down),
                "reset_input": Wire.to(SignalType.Boolean, reset),
            },
        ),
    )


def audio_switchbox(
    on_value: AudioInput = None,
    off_value: AudioInput = None,
    switch: BooleanInput = None,
) -> AudioWire:
    """
    Outputs the first input audio when receiving an on signal, and the second when receiving an off signal.
    """
    return Wire(
        SignalType.Audio,
        ComponentWrapper(
            AudioSwitchbox(0, Position()),
            {
                "on_value_input": Wire.to(SignalType.Audio, on_value),
                "off_value_input": Wire.to(SignalType.Audio, off_value),
                "switch_signal_input": Wire.to(SignalType.Boolean, switch),
            },
        ),
    )


def composite_switchbox(
    on_value: CompositeInput = None,
    off_value: CompositeInput = None,
    switch: BooleanInput = None,
) -> CompositeWire:
    """
    Outputs the first input composite when receiving an on signal, and the second when receiving an off signal.
    """
    return Wire(
        SignalType.Composite,
        ComponentWrapper(
            CompositeSwitchbox(0, Position()),
            {
                "on_value_input": Wire.to(SignalType.Composite, on_value),
                "off_value_input": Wire.to(SignalType.Composite, off_value),
                "switch_signal_input": Wire.to(SignalType.Boolean, switch),
            },
        ),
    )


def video_switchbox(
    on_value: VideoInput = None,
    off_value: VideoInput = None,
    switch: BooleanInput = None,
) -> VideoWire:
    """
    Outputs the first input video when receiving an on signal, and the second when receiving an off signal.
    """
    return Wire(
        SignalType.Video,
        ComponentWrapper(
            VideoSwitchbox(0, Position()),
            {
                "on_value_input": Wire.to(SignalType.Video, on_value),
                "off_value_input": Wire.to(SignalType.Video, off_value),
                "switch_signal_input": Wire.to(SignalType.Boolean, switch),
            },
        ),
    )


def composite_binary_to_number(value: CompositeInput = None) -> NumberWire:
    """
    Reads the on/off signals of a composite link and encodes them in the bits of an output number.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            CompositeBinaryToNumber(0, Position()),
            {"signal_to_convert_input": Wire.to(SignalType.Composite, value)},
        ),
    )


def number_to_composite_binary(value: NumberInput = None) -> CompositeWire:
    """
    Converts a number (rounded) to binary and outputs the bits as composite on/off signals.
    """
    return Wire(
        SignalType.Composite,
        ComponentWrapper(
            NumberToCompositeBinary(0, Position()),
            {"number_to_convert_input": Wire.to(SignalType.Number, value)},
        ),
    )


def composite_read_number(
    composite: CompositeInput = None,
    channel: int = 1,
    dynamic_channel: NumberInput = None,
) -> NumberWire:
    """
    Reads the number value from a selected channel of a composite input.
    If channel is 0, will read from the channel specified by the dynamic channel input.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            CompositeReadNumber(
                0, Position(), channel if dynamic_channel is None else 0
            ),
            {
                "composite_signal_input": Wire.to(SignalType.Composite, composite),
                "start_channel_input": Wire.to(SignalType.Number, dynamic_channel),
            },
        ),
    )


def composite_read_boolean(
    composite: CompositeInput = None,
    channel: int = 1,
    dynamic_channel: NumberInput = None,
) -> BooleanWire:
    """
    Reads the on/off value from a selected channel of a composite input.
    If channel is 0, will read from the channel specified by the dynamic channel input.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            CompositeReadBoolean(
                0, Position(), channel if dynamic_channel is None else 0
            ),
            {
                "composite_signal_input": Wire.to(SignalType.Composite, composite),
                "start_channel_input": Wire.to(SignalType.Number, dynamic_channel),
            },
        ),
    )


def composite_write_number(
    composite: CompositeInput = None,
    channel1: NumberInput = None,
    channel2: NumberInput = None,
    channel3: NumberInput = None,
    channel4: NumberInput = None,
    channel5: NumberInput = None,
    channel6: NumberInput = None,
    channel7: NumberInput = None,
    channel8: NumberInput = None,
    channel9: NumberInput = None,
    channel10: NumberInput = None,
    channel11: NumberInput = None,
    channel12: NumberInput = None,
    channel13: NumberInput = None,
    channel14: NumberInput = None,
    channel15: NumberInput = None,
    channel16: NumberInput = None,
    channel17: NumberInput = None,
    channel18: NumberInput = None,
    channel19: NumberInput = None,
    channel20: NumberInput = None,
    channel21: NumberInput = None,
    channel22: NumberInput = None,
    channel23: NumberInput = None,
    channel24: NumberInput = None,
    channel25: NumberInput = None,
    channel26: NumberInput = None,
    channel27: NumberInput = None,
    channel28: NumberInput = None,
    channel29: NumberInput = None,
    channel30: NumberInput = None,
    channel31: NumberInput = None,
    channel32: NumberInput = None,
    dynamic_channel: NumberInput = None,
) -> CompositeWire:
    """
    Outputs a composite signal that is the same as the input composite, but with one channel replaced by a number value.
    If channel is 0, will write to the channel specified by the dynamic channel input.
    """
    return Wire(
        SignalType.Composite,
        ComponentWrapper(
            CompositeWriteNumber(
                0, Position(), 1 if dynamic_channel is None else 0, 32
            ),
            {
                "composite_signal_input": Wire.to(SignalType.Composite, composite),
                "start_channel_input": Wire.to(SignalType.Number, dynamic_channel),
                "channel_1_input": Wire.to(SignalType.Number, channel1),
                "channel_2_input": Wire.to(SignalType.Number, channel2),
                "channel_3_input": Wire.to(SignalType.Number, channel3),
                "channel_4_input": Wire.to(SignalType.Number, channel4),
                "channel_5_input": Wire.to(SignalType.Number, channel5),
                "channel_6_input": Wire.to(SignalType.Number, channel6),
                "channel_7_input": Wire.to(SignalType.Number, channel7),
                "channel_8_input": Wire.to(SignalType.Number, channel8),
                "channel_9_input": Wire.to(SignalType.Number, channel9),
                "channel_10_input": Wire.to(SignalType.Number, channel10),
                "channel_11_input": Wire.to(SignalType.Number, channel11),
                "channel_12_input": Wire.to(SignalType.Number, channel12),
                "channel_13_input": Wire.to(SignalType.Number, channel13),
                "channel_14_input": Wire.to(SignalType.Number, channel14),
                "channel_15_input": Wire.to(SignalType.Number, channel15),
                "channel_16_input": Wire.to(SignalType.Number, channel16),
                "channel_17_input": Wire.to(SignalType.Number, channel17),
                "channel_18_input": Wire.to(SignalType.Number, channel18),
                "channel_19_input": Wire.to(SignalType.Number, channel19),
                "channel_20_input": Wire.to(SignalType.Number, channel20),
                "channel_21_input": Wire.to(SignalType.Number, channel21),
                "channel_22_input": Wire.to(SignalType.Number, channel22),
                "channel_23_input": Wire.to(SignalType.Number, channel23),
                "channel_24_input": Wire.to(SignalType.Number, channel24),
                "channel_25_input": Wire.to(SignalType.Number, channel25),
                "channel_26_input": Wire.to(SignalType.Number, channel26),
                "channel_27_input": Wire.to(SignalType.Number, channel27),
                "channel_28_input": Wire.to(SignalType.Number, channel28),
                "channel_29_input": Wire.to(SignalType.Number, channel29),
                "channel_30_input": Wire.to(SignalType.Number, channel30),
                "channel_31_input": Wire.to(SignalType.Number, channel31),
                "channel_32_input": Wire.to(SignalType.Number, channel32),
            },
        ),
    )


def composite_write_boolean(
    composite: CompositeInput = None,
    channel1: BooleanInput = None,
    channel2: BooleanInput = None,
    channel3: BooleanInput = None,
    channel4: BooleanInput = None,
    channel5: BooleanInput = None,
    channel6: BooleanInput = None,
    channel7: BooleanInput = None,
    channel8: BooleanInput = None,
    channel9: BooleanInput = None,
    channel10: BooleanInput = None,
    channel11: BooleanInput = None,
    channel12: BooleanInput = None,
    channel13: BooleanInput = None,
    channel14: BooleanInput = None,
    channel15: BooleanInput = None,
    channel16: BooleanInput = None,
    channel17: BooleanInput = None,
    channel18: BooleanInput = None,
    channel19: BooleanInput = None,
    channel20: BooleanInput = None,
    channel21: BooleanInput = None,
    channel22: BooleanInput = None,
    channel23: BooleanInput = None,
    channel24: BooleanInput = None,
    channel25: BooleanInput = None,
    channel26: BooleanInput = None,
    channel27: BooleanInput = None,
    channel28: BooleanInput = None,
    channel29: BooleanInput = None,
    channel30: BooleanInput = None,
    channel31: BooleanInput = None,
    channel32: BooleanInput = None,
    dynamic_channel: NumberInput = None,
) -> CompositeWire:
    """
    Outputs a composite signal that is the same as the input composite, but with one channel replaced by a boolean value.
    If channel is 0, will write to the channel specified by the dynamic channel input.
    """
    return Wire(
        SignalType.Composite,
        ComponentWrapper(
            CompositeWriteBoolean(
                0, Position(), 1 if dynamic_channel is None else 0, 32
            ),
            {
                "composite_signal_input": Wire.to(SignalType.Composite, composite),
                "start_channel_input": Wire.to(SignalType.Number, dynamic_channel),
                "channel_1_input": Wire.to(SignalType.Boolean, channel1),
                "channel_2_input": Wire.to(SignalType.Boolean, channel2),
                "channel_3_input": Wire.to(SignalType.Boolean, channel3),
                "channel_4_input": Wire.to(SignalType.Boolean, channel4),
                "channel_5_input": Wire.to(SignalType.Boolean, channel5),
                "channel_6_input": Wire.to(SignalType.Boolean, channel6),
                "channel_7_input": Wire.to(SignalType.Boolean, channel7),
                "channel_8_input": Wire.to(SignalType.Boolean, channel8),
                "channel_9_input": Wire.to(SignalType.Boolean, channel9),
                "channel_10_input": Wire.to(SignalType.Boolean, channel10),
                "channel_11_input": Wire.to(SignalType.Boolean, channel11),
                "channel_12_input": Wire.to(SignalType.Boolean, channel12),
                "channel_13_input": Wire.to(SignalType.Boolean, channel13),
                "channel_14_input": Wire.to(SignalType.Boolean, channel14),
                "channel_15_input": Wire.to(SignalType.Boolean, channel15),
                "channel_16_input": Wire.to(SignalType.Boolean, channel16),
                "channel_17_input": Wire.to(SignalType.Boolean, channel17),
                "channel_18_input": Wire.to(SignalType.Boolean, channel18),
                "channel_19_input": Wire.to(SignalType.Boolean, channel19),
                "channel_20_input": Wire.to(SignalType.Boolean, channel20),
                "channel_21_input": Wire.to(SignalType.Boolean, channel21),
                "channel_22_input": Wire.to(SignalType.Boolean, channel22),
                "channel_23_input": Wire.to(SignalType.Boolean, channel23),
                "channel_24_input": Wire.to(SignalType.Boolean, channel24),
                "channel_25_input": Wire.to(SignalType.Boolean, channel25),
                "channel_26_input": Wire.to(SignalType.Boolean, channel26),
                "channel_27_input": Wire.to(SignalType.Boolean, channel27),
                "channel_28_input": Wire.to(SignalType.Boolean, channel28),
                "channel_29_input": Wire.to(SignalType.Boolean, channel29),
                "channel_30_input": Wire.to(SignalType.Boolean, channel30),
                "channel_31_input": Wire.to(SignalType.Boolean, channel31),
                "channel_32_input": Wire.to(SignalType.Boolean, channel32),
            },
        ),
    )


def lua_script(
    script: str,
    composite_input: CompositeInput = None,
    video_input: VideoInput = None,
    minify: bool = True,
) -> tuple[CompositeWire, VideoWire]:
    """
    Executes a Lua script to process composite and video inputs and produce composite and video outputs.
    The script must define two functions: processComposite(input) and processVideo(input).
    """
    script = script_handling.verify_script(script, minify)
    component = ComponentWrapper(
        LuaScript(0, Position(), script),
        {
            "data_input": Wire.to(SignalType.Composite, composite_input),
            "video_input": Wire.to(SignalType.Video, video_input),
        },
    )
    return Wire(SignalType.Composite, component), Wire(
        SignalType.Video, component, node_index=1
    )


def lua_script_file(
    file_path: str,
    composite_input: CompositeInput = None,
    video_input: VideoInput = None,
    minify: bool = True,
) -> tuple[CompositeWire, VideoWire]:
    """
    Takes a lua script from a file and embeds it. The file may use dependencies, which are automatically resolved.
    See lua_script for more information.
    """
    return lua_script(
        script_handling.resolve_and_verify_script(file_path),
        composite_input,
        video_input,
        minify,
    )


def property_dropdown(
    options: dict[str, Number], default: str, name: str = "value"
) -> NumberWire:
    """
    A dropdown property that can be set in the microcontroller properties.
    Options should be a dictionary mapping option names to their corresponding numeric values.
    The default parameter should be the name of the option that is selected by default.
    """
    if default not in options:
        raise ValueError(f"Default value '{default}' not in options")
    dropdown_options = [DropDownOption(name, value) for name, value in options.items()]
    default_idx: int = list(options.keys()).index(default)
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            PropertyDropdown(0, Position(), default_idx, dropdown_options, name), {}
        ),
    )


def property_number(value: Number, name: str = "value") -> NumberWire:
    """
    A number property that can be set in the microcontroller properties.
    """
    return Wire(
        SignalType.Number,
        ComponentWrapper(PropertyNumber(0, Position(), name, value), {}),
    )


def property_slider(
    min_value: Number = 0,
    max_value: Number = 10,
    rounding: Number = 1,
    value: Number = 0,
    name: str = "value",
) -> NumberWire:
    """
    A slider property that can be set in the microcontroller properties.
    """
    if not min_value <= value <= max_value:
        raise ValueError("Value must be between min and max values")
    return Wire(
        SignalType.Number,
        ComponentWrapper(
            PropertySlider(0, Position(), min_value, max_value, value, rounding, name),
            {},
        ),
    )


def property_toggle(
    value: bool = False,
    name: str = "value",
    on_label: str = "on",
    off_label: str = "off",
) -> BooleanWire:
    """
    A toggle property that can be set in the microcontroller properties.
    """
    return Wire(
        SignalType.Boolean,
        ComponentWrapper(
            PropertyToggle(0, Position(), name, on_label, off_label, value), {}
        ),
    )



from sw_mc_lib.Types import PulseMode, SignalType

import sw_mc_builder.components as comp
from sw_mc_builder import util
from sw_mc_builder._utils import add_include_path
from sw_mc_builder.handling import handle_mcs
from sw_mc_builder.microcontroller import Microcontroller
from sw_mc_builder.wire import (
    AudioWire,
    BooleanWire,
    CompositeWire,
    NumberWire,
    VideoWire,
)

__all__ = [
    "add_include_path",
    "comp",
    "util",
    "handle_mcs",
    "Microcontroller",
    "SignalType",
    "PulseMode",
    "NumberWire",
    "BooleanWire",
    "CompositeWire",
    "AudioWire",
    "VideoWire",
]



import sys
from pathlib import Path

import tumfl
from sw_mc_lib import Position
from sw_mc_lib.Components import (
    PropertyDropdown,
    PropertyNumber,
    PropertySlider,
    PropertyText,
    PropertyToggle,
)

MAIN_PATH: Path = Path(sys.argv[0]).parent.resolve()

# use newline as statement separator, since it is more readable and does not consume more space
tumfl.formatter.MinifiedStyle.STATEMENT_SEPARATOR = "\n"

PROPERTIES = (
    PropertyNumber,
    PropertySlider,
    PropertyText,
    PropertyToggle,
    PropertyDropdown,
)
PROPERTY_IDS: set[str] = {str(cls(0, Position()).type.value) for cls in PROPERTIES}

BUILDER_IDENTIFIER: str = "Built with sw_mc_builder version 1"

INCLUDE_PATHS: list[Path] = [MAIN_PATH]


def normalize_path(path: str | Path) -> Path:
    """
    Normalize a path by resolving it and converting it to an absolute path.
    """
    if (MAIN_PATH / path).exists():
        return (MAIN_PATH / path).resolve()
    for include_path in INCLUDE_PATHS:
        combined_path = (include_path / path).resolve()
        if combined_path.exists():
            return combined_path
    raise FileNotFoundError(f"Could not find path: {path}")


def add_include_path(path: str | Path) -> None:
    """
    Add a path to the list of include paths for resolving scripts and script dependencies.
    """
    normalized_path = normalize_path(path)
    if normalized_path not in INCLUDE_PATHS:
        INCLUDE_PATHS.append(normalized_path)



import unittest

from sw_mc_lib import Position
from sw_mc_lib.Components import ConstantOn
from sw_mc_lib.Types import SignalType

from sw_mc_builder.component_wrapper import ComponentWrapper
from sw_mc_builder.wire import Wire


class Test(unittest.TestCase):
    def test_placeholder(self) -> None:
        Wire(SignalType.Video, ComponentWrapper(ConstantOn(1, Position()), {}))



# pylint: disable=too-few-public-methods
from __future__ import annotations


class PseudoComponent:
    """A pseudo component acing as a marker for the builder. Not a subclass of Component"""


class Unconnected(PseudoComponent):
    """A constant component that outputs nothing and has no size. Not a subclass of Placeholder"""


class Placeholder(PseudoComponent):
    """A placeholder component that does nothing and has no size."""


class InputPlaceholder(PseudoComponent):
    """A placeholder component that represents an input to the microcontroller."""

    def __init__(self, component_id: int, name: str, description: str) -> None:
        self.name: str = name
        self.component_id: int = component_id
        self.description: str = description



from __future__ import annotations

from sw_mc_lib.Types import SignalType

import sw_mc_builder.components as comp
from sw_mc_builder import wire as wire_


def bool_to_int(
    value: wire_.BooleanWire, true_value: int = 1, false_value: int = 0
) -> wire_.NumberWire:
    return comp.numerical_switchbox(true_value, false_value, value)


def moving_avg(wire: wire_.NumberWire, count: int = 60) -> wire_.NumberWire:
    """
    Simple moving average over `count` ticks.
    :param wire: The input wire.
    :param count: Number of ticks to average over.
    :return: The averaged wire.
    """
    end = wire
    for _ in range(count):
        end = comp.function("x", end).stop_optimization()
    result = comp.placeholder(SignalType.Number)
    result.replace_producer(result + wire - end)
    return result / count



import warnings
from pathlib import Path
from typing import Optional

from sw_mc_lib import Node, NodePosition, Position, XMLParserElement
from sw_mc_lib.Components import (
    PropertyDropdown,
    PropertyNumber,
    PropertySlider,
    PropertyText,
    PropertyToggle,
    TooltipBoolean,
    TooltipNumber,
)
from sw_mc_lib.layout import layout_mc
from sw_mc_lib.Microcontroller import MCImage
from sw_mc_lib.Microcontroller import Microcontroller as SWMicrocontroller
from sw_mc_lib.optimizer import (
    optimize_composite_writes,
    optimize_functions,
    optimize_tree,
)
from sw_mc_lib.Types import NodeMode, SignalType, TooltipMode

from sw_mc_builder._utils import normalize_path
from sw_mc_builder.component_wrapper import ComponentWrapper
from sw_mc_builder.pseudo_components import InputPlaceholder, PseudoComponent
from sw_mc_builder.wire import BooleanInput, NumberInput, Wire

from ._utils import BUILDER_IDENTIFIER, PROPERTIES
from .optimizer import optimize_arithmetic


class Microcontroller:
    def __init__(
        self,
        name: str,
        width: int = 2,
        length: int = 2,
        description: str = "No description set.",
        save_name: Optional[str] = None,
    ):
        self._mc: SWMicrocontroller = SWMicrocontroller(
            name, description, width, length, [], []
        )
        self._save_name: str = save_name or name
        self._placed_inputs: set[InputPlaceholder] = set()
        self.optimize: bool = True
        self._placed_outputs: list[tuple[Wire, Node]] = []
        self._additional_components: list[ComponentWrapper] = []
        self._warned_placement: bool = False
        self._properties: set[ComponentWrapper] = set()

    def stop_optimization(self) -> None:
        self.optimize = False

    def add_image_from_file(self, file: str | Path) -> None:
        self._mc.image = MCImage.from_png(normalize_path(file))

    def add_image_from_list(self, pixels: list[list[bool]]) -> None:
        self._mc.image = MCImage(pixels)

    def save_image(self, path: Path) -> None:
        self._mc.image.to_png(path)

    def _validate_placement(self, position: NodePosition) -> None:
        for node in self._mc.nodes:
            if node.position == position:
                raise ValueError(f"Node already exists at position {position}")
        if position.x >= 6 or position.z >= 6 or position.x < 0 or position.z < 0:
            raise ValueError(f"Node position out of bounds {position}")
        if position.x >= self._mc.width or position.z >= self._mc.length:
            if not self._warned_placement:
                warnings.warn(
                    f"Warning: Placing node at {position} which is outside the microcontroller bounds ({self._mc.width}, {self._mc.length}). Microcontroller will be expanded.",
                    stacklevel=3,
                )
                self._warned_placement = True
            self._mc.width = max(self._mc.width, position.x + 1)
            self._mc.length = max(self._mc.length, position.z + 1)

    def place_input(self, input_: Wire, x: int = 0, y: int = 0) -> None:
        if not isinstance(input_.producer, InputPlaceholder):
            raise TypeError("Input must be produced by an InputPlaceholder")
        if input_.producer in self._placed_inputs:
            raise ValueError(f"Input {input_.producer.name} already placed")
        position: NodePosition = NodePosition(x, y)
        self._validate_placement(position)
        self._mc.add_new_node(
            Node(
                0,
                0,
                input_.producer.name,
                NodeMode.Input,
                input_.wire_type,
                input_.producer.description,
                NodePosition(x, y),
            )
        )
        input_.producer.component_id = self._mc.nodes[-1].component_id
        self._placed_inputs.add(input_.producer)

    def place_output(
        self,
        input_: Wire,
        name: str,
        description: str = "The processed output signal.",
        x: int = 0,
        y: int = 0,
    ) -> None:
        position: NodePosition = NodePosition(x, y)
        self._validate_placement(position)
        self._mc.add_new_node(
            Node(
                0,
                0,
                name,
                NodeMode.Output,
                input_.wire_type,
                description,
                NodePosition(x, y),
                input=input_.to_input(),
            )
        )
        self._placed_outputs.append((input_, self._mc.nodes[-1]))

    def add_number_tooltip(
        self,
        name: str = "value",
        display_number: NumberInput = None,
        is_error: BooleanInput = None,
        display: TooltipMode = TooltipMode.Always,
    ) -> None:
        self._additional_components.append(
            ComponentWrapper(
                TooltipNumber(-1, Position(0, 0), name, display),
                {
                    "display_number_input": Wire.to(SignalType.Number, display_number),
                    "is_error_input": Wire.to(SignalType.Boolean, is_error),
                },
            )
        )

    def add_boolean_tooltip(
        self,
        name: str = "value",
        value: BooleanInput = None,
        on_label: str = "on",
        off_label: str = "off",
        display: TooltipMode = TooltipMode.Always,
    ) -> None:
        self._additional_components.append(
            ComponentWrapper(
                TooltipBoolean(-1, Position(0, 0), name, on_label, off_label, display),
                {
                    "display_number_input": Wire.to(SignalType.Boolean, value),
                },
            )
        )

    def add_property(self, property: Wire) -> None:
        if not isinstance(property.producer, ComponentWrapper):
            raise TypeError("Property must be produced by a ComponentWrapper")
        if not isinstance(
            property.producer.inner_component,
            (PropertyNumber, PropertySlider, PropertyToggle, PropertyDropdown),
        ):
            raise TypeError("Property must be a valid property component")
        self._additional_components.append(property.producer)

    def add_text_property(
        self, name: str, content: str, force_property: bool = False
    ) -> None:
        component = ComponentWrapper(
            PropertyText(-1, Position(0, 0), name, content),
            {},
        )
        component.force_property = force_property
        self._additional_components.append(component)

    def _resolve(self) -> None:
        to_visit: set[ComponentWrapper | PseudoComponent] = set()
        visited_nodes: set[ComponentWrapper | PseudoComponent] = set()
        visited_components: set[ComponentWrapper] = set()
        next_id: int = self._mc.id_counter + 1
        for node_source, _ in self._placed_outputs:
            to_visit.add(node_source.producer)

        for component in self._additional_components:
            # This is done separately in order for the order to be represented in the final microcontroller
            # This is especially relevant for tooltips (as they are shown in order)
            for input_wire in component.inputs.values():
                to_visit.add(input_wire.producer)
            visited_components.add(component)
            if isinstance(component.inner_component, PROPERTIES):
                self._properties.add(component)
            component.component_id = next_id
            next_id += 1

        while to_visit:
            current = to_visit.pop()
            if current in visited_nodes:
                continue
            visited_nodes.add(current)

            if isinstance(current, ComponentWrapper):
                for input_wire in current.inputs.values():
                    to_visit.add(input_wire.producer)
                visited_components.add(current)
                if isinstance(current.inner_component, PROPERTIES):
                    self._properties.add(current)
                current.component_id = next_id
                next_id += 1
            elif isinstance(current, InputPlaceholder):
                if current not in self._placed_inputs:
                    raise ValueError(
                        f"Input {current.name} not placed in microcontroller"
                    )

        for wire, node in self._placed_outputs:
            node.input = wire.to_input()
            if node.input:
                node.input.index = "1"

        for component in visited_components:
            self._mc.components.append(component.to_component())

    def _resolve_and_optimize(self) -> XMLParserElement:
        if self.optimize:
            wires = [
                wire
                for component in self._additional_components
                for wire in component.inputs.values()
            ]
            for node, _ in self._placed_outputs:
                wires.append(node)
            optimize_arithmetic(wires)
        self._resolve()
        optimize_composite_writes(self._mc)
        optimize_functions(self._mc)
        optimize_tree(self._mc)
        layout_mc(self._mc)
        mc_xml = self._mc.to_xml()

        group = mc_xml.children[1]
        assert group.tag == "group"
        data = group.children[0]
        assert data.tag == "data"
        data.attributes["desc"] = BUILDER_IDENTIFIER

        properties = {
            str(prop.component_id): prop
            for prop in self._properties
            if prop.force_property
        }
        components = group.children[1]
        assert components.tag == "components"
        for component in components.children:
            obj = component.children[0]
            assert obj.tag == "object"
            if obj.attributes.get("id", "0") in properties:
                obj.attributes["force_property"] = "1"

        return mc_xml



from __future__ import annotations

import builtins
import inspect
import re
import warnings
from typing import Generic, Iterable, Literal, Optional, TypeVar, overload

from sw_mc_lib import Input, Position
from sw_mc_lib.Components import ConstantNumber, ConstantOn
from sw_mc_lib.Types import SignalType

import sw_mc_builder.components as comp
from sw_mc_builder import util

from .component_wrapper import ComponentWrapper
from .pseudo_components import (
    InputPlaceholder,
    Placeholder,
    PseudoComponent,
    Unconnected,
)

T = TypeVar("T", bound=SignalType)


class Wire(Generic[T]):
    # pylint: disable=too-many-public-methods

    SET_VIA_SET_METHOD: set[ComponentWrapper | PseudoComponent] = set()

    def __init__(
        self,
        wire_type: T,
        producer: ComponentWrapper | PseudoComponent,
        node_index: int = 0,
    ):
        self.wire_type: SignalType = wire_type
        self.producer: ComponentWrapper | PseudoComponent = producer
        self.node_index: int = node_index

    def stop_optimization(self) -> Wire[T]:
        """Will prevent the optimizer from removing this component or merging it with other components."""
        if isinstance(self.producer, ComponentWrapper):
            self.producer.optimize = False
        return self

    def force_property(self) -> Wire[T]:
        """Will overwrite the contents of a property, even if that property is already defined in a vehicle."""
        if isinstance(self.producer, ComponentWrapper):
            self.producer.force_property = True
        return self

    @staticmethod
    @overload
    def to(
        signal_type: Literal[SignalType.Number], wire: NumberInput
    ) -> NumberWire: ...

    @staticmethod
    @overload
    def to(
        signal_type: Literal[SignalType.Boolean], wire: BooleanInput
    ) -> BooleanWire: ...

    @staticmethod
    @overload
    def to(
        signal_type: Literal[SignalType.Composite], wire: CompositeInput
    ) -> CompositeWire: ...

    @staticmethod
    @overload
    def to(signal_type: Literal[SignalType.Audio], wire: AudioInput) -> AudioWire: ...

    @staticmethod
    @overload
    def to(signal_type: Literal[SignalType.Video], wire: VideoInput) -> VideoWire: ...

    @staticmethod
    def to(signal_type: SignalType, wire: Wire | int | float | bool | None) -> Wire:
        if isinstance(wire, Wire):
            if wire.wire_type != signal_type:
                raise TypeError(
                    f"Wire type mismatch: expected {signal_type}, got {wire.wire_type}"
                )
            return wire
        if signal_type == SignalType.Boolean and isinstance(wire, bool):
            if wire:
                return Wire(
                    SignalType.Boolean, ComponentWrapper(ConstantOn(1, Position()), {})
                )
            return Wire(SignalType.Boolean, Unconnected())
        if signal_type == SignalType.Number and isinstance(wire, (int, float)):
            return Wire(
                SignalType.Number,
                ComponentWrapper(ConstantNumber(0, Position(), wire), {}),
            )
        if wire is None:
            return Wire(signal_type, Unconnected())
        raise TypeError(f"Wire type mismatch: expected {signal_type}, got {wire}")

    def replace_producer(self, new_wire: Wire[T]) -> None:
        if self.wire_type != new_wire.wire_type:
            raise TypeError(
                f"Wire type mismatch: expected {self.wire_type}, got {new_wire.wire_type}"
            )
        self.producer = new_wire.producer

    @property
    def component_id(self) -> Optional[int]:
        if isinstance(self.producer, (ComponentWrapper, InputPlaceholder)):
            return self.producer.component_id
        if isinstance(self.producer, Unconnected):
            return None
        if isinstance(self.producer, Placeholder):
            raise RuntimeError(
                "Placeholder needs to be replaced before compiling microcontroller"
            )
        raise TypeError(f"Invalid producer type: {type(self.producer)}")

    def __hash__(self) -> int:
        return hash(id(self))

    def to_input(self) -> Optional[Input]:
        if self.component_id is None:
            return None
        return Input(self.component_id, node_index=self.node_index)

    def __add__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only add NumberWire")
        return comp.add(self, other)  # type: ignore[arg-type]

    def __sub__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only subtract NumberWire")
        return comp.sub(self, other)  # type: ignore[arg-type]

    def __mul__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only multiply NumberWire")
        return comp.mul(self, other)  # type: ignore[arg-type]

    def __truediv__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only divide NumberWire")
        return comp.div(self, other)[0]  # type: ignore[arg-type]

    def __mod__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only modulo NumberWire")
        return comp.mod(self, other)  # type: ignore[arg-type]

    def __abs__(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only abs NumberWire")
        return comp.abs(self)  # type: ignore[arg-type]

    def __lshift__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only left shift NumberWire")
        return comp.function("x*2^y", self, other)  # type: ignore[arg-type]

    def __rshift__(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only right shift NumberWire")
        return comp.function("x/2^y", self, other)  # type: ignore[arg-type]

    def __neg__(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only negate NumberWire")
        return comp.sub(comp.unconnected(SignalType.Number), self)  # type: ignore[arg-type]

    def __pow__(self, power: NumberInput, modulo: NumberInput = None) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only exponentiate NumberWire")
        if modulo is not None:
            return comp.function("(x^y)%z", self, power, modulo)  # type: ignore[arg-type]
        return comp.function("x^y", self, power)  # type: ignore[arg-type]

    @staticmethod
    def __slice_to_range(s: object) -> list[int]:
        if isinstance(s, slice):
            start = s.start if s.start is not None else 1
            stop = s.stop if s.stop is not None else 32
            step = s.step if s.step is not None else 1
            if not (1 <= start <= 32 and 1 <= stop <= 32):
                raise ValueError("Slice indices must be in the range [1,32]")
            if step == 0:
                raise ValueError("Slice step cannot be zero")
            return list(range(start, stop + 1, step))
        if isinstance(s, int):
            if not 1 <= s <= 32:
                raise ValueError("Index must be in the range [1,32]")
            return [s]
        raise TypeError("Index must be an integer or slice")

    @overload
    def __getitem__(self, item: int) -> NumberWire: ...

    @overload
    def __getitem__(self, item: slice) -> list[NumberWire]: ...

    def __getitem__(self, item: int | slice) -> NumberWire | list[NumberWire]:
        if self.wire_type != SignalType.Composite:
            raise TypeError("Can only index CompositeWire")
        index_slice = self.__slice_to_range(item)
        if isinstance(item, slice):
            return [
                comp.composite_read_number(self, idx)  # type: ignore[arg-type]
                for idx in index_slice
            ]
        return comp.composite_read_number(self, item)  # type: ignore[arg-type]

    @overload
    def get_bool(self, item: int) -> BooleanWire: ...

    @overload
    def get_bool(self, item: slice) -> list[BooleanWire]: ...

    def get_bool(self, item: int | slice) -> BooleanWire | list[BooleanWire]:
        if self.wire_type != SignalType.Composite:
            raise TypeError("Can only index CompositeWire")
        index_slice = self.__slice_to_range(item)
        if isinstance(item, slice):
            return [
                comp.composite_read_boolean(self, idx)  # type: ignore[arg-type]
                for idx in index_slice
            ]
        return comp.composite_read_boolean(self, item)  # type: ignore[arg-type]

    @staticmethod
    def __is_bool(element: object) -> bool:
        return isinstance(element, bool) or (
            isinstance(element, Wire) and element.wire_type == SignalType.Boolean
        )

    @staticmethod
    def __is_number(element: object) -> bool:
        return isinstance(element, (int, float)) or (
            isinstance(element, Wire) and element.wire_type == SignalType.Number
        )

    @overload
    def set(self, item: int, value: NumberInput | BooleanInput) -> CompositeWire: ...

    @overload
    def set(
        self, item: slice, value: Iterable[NumberWire | int | float]
    ) -> CompositeWire: ...

    @overload
    def set(
        self, item: slice, value: Iterable[BooleanWire | bool]
    ) -> CompositeWire: ...

    def set(
        self,
        item: int | slice,
        value: (
            NumberInput
            | BooleanInput
            | Iterable[NumberWire | int | float]
            | Iterable[BooleanWire | bool]
        ),
    ) -> CompositeWire:
        """Alias for __setitem__ to allow using .set() instead of []= for setting composite values."""
        self.SET_VIA_SET_METHOD.add(self.producer)
        if self.wire_type != SignalType.Composite:
            raise TypeError("Can only index CompositeWire")
        index_slice = self.__slice_to_range(item)
        if isinstance(item, slice):
            if not isinstance(value, Iterable):
                raise TypeError("Value must be an iterable when using slice assignment")
            value_list = list(value)
            if len(value_list) != len(index_slice):
                raise ValueError("Value length must match slice length")
            if len(index_slice) == 0:
                return self  # type: ignore[return-value]
            composite_write = (
                comp.composite_write_number
                if self.__is_number(value_list[0])
                else comp.composite_write_boolean
            )
            return composite_write(self, **{f"channel{idx}": val for idx, val in zip(index_slice, value_list)})  # type: ignore[arg-type]
        if self.__is_number(value):
            return comp.composite_write_number(self, **{f"channel{item}": value})  # type: ignore[arg-type]
        if self.__is_bool(value):
            return comp.composite_write_boolean(self, **{f"channel{item}": value})  # type: ignore[arg-type]
        raise TypeError("Can only write NumberWire or BooleanWire to CompositeWire")

    @overload
    def __setitem__(self, item: int, value: NumberInput | BooleanInput) -> None: ...

    @overload
    def __setitem__(
        self, item: slice, value: Iterable[NumberWire | int | float]
    ) -> None: ...

    @overload
    def __setitem__(self, item: slice, value: Iterable[BooleanWire | bool]) -> None: ...

    def __setitem__(
        self,
        item: int | slice,
        value: (
            NumberInput
            | BooleanInput
            | Iterable[NumberWire | int | float]
            | Iterable[BooleanWire | bool]
        ),
    ) -> None:
        if self.producer in self.SET_VIA_SET_METHOD:
            warnings.warn(
                "Using both `wire[index] = value` and `wire.set(index, value)` on the same CompositeWire is not recommended. This may lead to surprising behaviour.",
                stacklevel=2,
            )
            stack = inspect.stack()
            if len(stack) > 1:
                frame_info = stack[1]
                context = frame_info.code_context
                index = frame_info.index
                if context is not None and index is not None:
                    recommendation = re.sub(
                        r"^([a-zA-Z_][a-zA-Z0-9_]*)\s*\[([^]]+)]\s*=\s*(.*)$",
                        r"\1 = \1.set(\2, \3)",
                        context[index],
                    )
                    if recommendation != context[index]:
                        print(f"Recommended replacement:\n  {recommendation.strip()}")
        clone = Wire(self.wire_type, self.producer, self.node_index)
        self.replace_producer(clone.set(item, value))  # type: ignore[arg-type]

    def __eq__(self, other: object) -> BooleanWire:  # type: ignore[override]
        if not isinstance(other, (Wire, int, float)):
            return NotImplemented
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.equal(self, other)  # type: ignore[arg-type]

    def __ne__(self, other: object) -> BooleanWire:  # type: ignore[override]
        if not isinstance(other, (Wire, int, float)):
            return NotImplemented
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.not_(comp.equal(self, other))  # type: ignore[arg-type]

    def __lt__(self, other: NumberInput) -> BooleanWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.less_than(self, other)  # type: ignore[arg-type]

    def __le__(self, other: NumberInput) -> BooleanWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.or_(comp.less_than(self, other), comp.equal(self, other))  # type: ignore[arg-type]

    def __gt__(self, other: NumberInput) -> BooleanWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.greater_than(self, other)  # type: ignore[arg-type]

    def __ge__(self, other: NumberInput) -> BooleanWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only compare NumberWire")
        return comp.or_(comp.greater_than(self, other), comp.equal(self, other))  # type: ignore[arg-type]

    def __and__(self, other: BooleanInput) -> BooleanWire:
        if self.wire_type != SignalType.Boolean:
            raise TypeError("Can only AND BooleanWire")
        return comp.and_(self, other)  # type: ignore[arg-type]

    def __or__(self, other: BooleanInput) -> BooleanWire:
        if self.wire_type != SignalType.Boolean:
            raise TypeError("Can only OR BooleanWire")
        return comp.or_(self, other)  # type: ignore[arg-type]

    def __invert__(self) -> BooleanWire:
        if self.wire_type != SignalType.Boolean:
            raise TypeError("Can only NOT BooleanWire")
        return comp.not_(self)  # type: ignore[arg-type]

    def int(self) -> NumberWire:
        if self.wire_type not in (SignalType.Number, SignalType.Boolean):
            raise TypeError("Can only convert NumberWire or BooleanWire to int")
        if self.wire_type == SignalType.Boolean:
            return util.bool_to_int(self)  # type: ignore[arg-type]
        return self  # type: ignore[return-value]

    def max(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only max NumberWire")
        return comp.function("max(x,y)", self, other)  # type: ignore[arg-type]

    def min(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only min NumberWire")
        return comp.function("min(x,y)", self, other)  # type: ignore[arg-type]

    def clamp(self, min_: NumberInput, max_: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only clamp NumberWire")
        return comp.function("clamp(x,y,z)", self, min_, max_)  # type: ignore[arg-type]

    def lerp(self, a: NumberInput, b: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only lerp NumberWire")
        return comp.function("lerp(x,y,z)", self, a, b)  # type: ignore[arg-type]

    def sin(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only sin NumberWire")
        return comp.function("sin(x)", self)  # type: ignore[arg-type]

    def cos(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only cos NumberWire")
        return comp.function("cos(x)", self)  # type: ignore[arg-type]

    def tan(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only tan NumberWire")
        return comp.function("tan(x)", self)  # type: ignore[arg-type]

    def asin(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only asin NumberWire")
        return comp.function("asin(x)", self)  # type: ignore[arg-type]

    def acos(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only acos NumberWire")
        return comp.function("acos(x)", self)  # type: ignore[arg-type]

    def atan(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only atan NumberWire")
        return comp.function("atan(x)", self)  # type: ignore[arg-type]

    def atan2(self, other: NumberInput) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only atan2 NumberWire")
        return comp.function("atan2(x,y)", self, other)  # type: ignore[arg-type]

    def ceil(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only ceil NumberWire")
        return comp.function("ceil(x)", self)  # type: ignore[arg-type]

    def floor(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only floor NumberWire")
        return comp.function("floor(x)", self)  # type: ignore[arg-type]

    def round(self, ndigits: NumberInput = 0) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only round NumberWire")
        return comp.function("round(x,y)", self, ndigits)  # type: ignore[arg-type]

    def sgn(self) -> NumberWire:
        """Sign function: -1 if x < 0, 1 if x == 0, 1 if x > 0"""
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only sgn NumberWire")
        return comp.function("sgn(x)", self)  # type: ignore[arg-type]

    def sqrt(self) -> NumberWire:
        if self.wire_type != SignalType.Number:
            raise TypeError("Can only sqrt NumberWire")
        return comp.function("sqrt(x)", self)  # type: ignore[arg-type]

    # These overloads are a mess, but they are the most precise I can make them

    @overload
    def switch(
        self,
        on_value: NumberWire,
        off_value: NumberWire | float | builtins.int,
    ) -> NumberWire: ...

    @overload
    def switch(
        self,
        on_value: NumberWire | float | builtins.int,
        off_value: NumberWire,
    ) -> NumberWire: ...

    @overload
    def switch(
        self, on_value: BooleanWire, off_value: BooleanWire | bool
    ) -> BooleanWire: ...

    @overload
    def switch(
        self, on_value: BooleanWire | bool, off_value: BooleanWire
    ) -> BooleanWire: ...

    # There is no way to tell a literal bool from a literal int in overloads.
    # They will be treated correctly in the implementation though.
    @overload
    def switch(
        self, on_value: bool | builtins.int, off_value: bool | builtins.int
    ) -> BooleanWire | NumberWire: ...

    @overload
    def switch(
        self, on_value: CompositeWire, off_value: CompositeWire
    ) -> CompositeWire: ...

    @overload
    def switch(self, on_value: AudioWire, off_value: AudioWire) -> AudioWire: ...

    @overload
    def switch(self, on_value: VideoWire, off_value: VideoWire) -> VideoWire: ...

    def switch(
        self,
        on_value: (
            NumberWire
            | float
            | builtins.int
            | BooleanWire
            | bool
            | CompositeWire
            | AudioWire
            | VideoWire
        ),
        off_value: (
            NumberWire
            | float
            | builtins.int
            | BooleanWire
            | bool
            | CompositeWire
            | AudioWire
            | VideoWire
        ),
    ) -> NumberWire | BooleanWire | CompositeWire | AudioWire | VideoWire:
        if self.wire_type != SignalType.Boolean:
            raise TypeError("Can only switch BooleanWire")
        if self.__is_bool(on_value):
            return comp.composite_switchbox(on_value, off_value, self)  # type: ignore[arg-type]
        if self.__is_number(on_value):
            return comp.numerical_switchbox(on_value, off_value, self)  # type: ignore[arg-type]
        assert isinstance(on_value, Wire)
        if on_value.wire_type == SignalType.Audio:
            return comp.audio_switchbox(on_value, off_value, self)  # type: ignore[arg-type]
        if on_value.wire_type == SignalType.Video:
            return comp.video_switchbox(on_value, off_value, self)  # type: ignore[arg-type]
        if on_value.wire_type == SignalType.Boolean:
            return comp.or_(comp.and_(self, on_value), comp.and_(comp.not_(self), off_value))  # type: ignore[arg-type]
        if on_value.wire_type == SignalType.Composite:
            return comp.composite_switchbox(on_value, off_value, self)  # type: ignore[arg-type]
        raise NotImplementedError(f"Switch on {on_value} not implemented")

    def __repr__(self) -> str:
        return f"Wire(type={self.wire_type}, producer={self.producer}, node_index={self.node_index})"


NumberWire = Wire[Literal[SignalType.Number]]
BooleanWire = Wire[Literal[SignalType.Boolean]]
CompositeWire = Wire[Literal[SignalType.Composite]]
AudioWire = Wire[Literal[SignalType.Audio]]
VideoWire = Wire[Literal[SignalType.Video]]

NumberInput = Wire[Literal[SignalType.Number]] | int | float | None
BooleanInput = Wire[Literal[SignalType.Boolean]] | bool | None
CompositeInput = Wire[Literal[SignalType.Composite]] | None
AudioInput = Wire[Literal[SignalType.Audio]] | None
VideoInput = Wire[Literal[SignalType.Video]] | None
